/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include "tfio.h"
#include <fcntl.h>
#include <unistd.h>
#ifndef _OS_WIN32
#include <sys/ioctl.h>
#include <linux/fs.h>
#endif

#define DEFAULT_IOfkt (IOfkt_CanWrite | IOfkt_CanSave)

typedef struct TWRSeg {
  struct TWRSeg *next;
  int BlockSize;
  int64_t BlockAdr;
  char data[];
} TWRSeg;


#define TCSelf TFIOfd
typedef struct TFIOfd {
  AlgTFIO
  TglobPath *gp;
  Tstat64 *st;
  FileHandle fd;
  int64_t size;
  int64_t Offset;
  int BlockShift;
  int BlockSize;
  int BlockCount;
  int BlockMemSize;
  char *BlockBuf;
  int64_t BlockAdr;
  TWRSeg *wrseg;
  bool IsOpen;
} TFIOfd;
#undef TCSelf

#define MAX_MEM 0x10000

static bool TFIOfd_seek(TFIOfd *self, int64_t Offset) {
	net_err_clr(&self->ErrRet);
	if (file_seek(self->fd, &self->ErrRet, Offset)) return True;
	net_err_app_4(&self->ErrRet, self->FileName, "seek", EMSG_ERR1);
	return False;
}

static bool TFIOfd_read(TFIOfd *self, void *buf, size_t bufsize) {
	net_err_clr(&self->ErrRet);
	if (file_read(self->fd, &self->ErrRet, buf, bufsize, NULL)) return True;
	net_err_app_4(&self->ErrRet, self->FileName, "read", EMSG_ERR1);
	return False;
}

static int64_t GetBlockAdr(TFIOfd *self, int64_t Offset) {
	return Offset & ~((1 << self->BlockShift) - 1);
}


static TWRSeg *AddWRSeg(TFIOfd *self, int64_t Offset) {
  TWRSeg *wrseg, **ppws;
  int64_t off;

	wrseg = xmalloc(sizeof(TWRSeg) + self->BlockSize);
	wrseg->next = NULL;
	wrseg->BlockSize = self->BlockSize;
	wrseg->BlockAdr = GetBlockAdr(self, Offset);
	off = wrseg->BlockAdr - self->BlockAdr;
	if ((off >= 0) && ((off + self->BlockSize) <= self->BlockMemSize)) {
	    pnmove(wrseg->data, self->BlockBuf + off, self->BlockSize);
	} else {
	    net_err_clr(&self->ErrRet);
	    if (!TFIOfd_seek(self, wrseg->BlockAdr) ||
	        !TFIOfd_read(self, wrseg->data, wrseg->BlockSize)) {
		xfree(wrseg);
		return NULL;
	    }
	}

	ppws = &self->wrseg;
	while (*ppws) ppws = &(*ppws)->next;
	*ppws = wrseg;
	return wrseg;
}

static int64_t fio_GetSizeAbs(TFIOfd *self) {
	return self->size;
}

static int64_t TFIOfd_GetSize(TFIOfd *self) {
	return fio_GetSizeAbs(self) - self->Offset;
}

static bool fio_SetOffset(TFIOfd *self, int64_t Offset) {
	if (Offset < 0) Offset = 0;
	if (Offset > self->size) Offset = self->size;
	if (self->Offset != Offset) {
	    self->Offset = Offset;
	    if (self->OnSetOffset) self->OnSetOffset(self, self->Offset);
	}
	return True;
}

static int64_t fio_GetOffset(TFIOfd *self) {
	return self->Offset;
}

static char *TFIOfd_GetDataP(TFIOfd *self, int64_t Offset, int MaxChr) {
  bool rdseg;

	net_err_clr(&self->ErrRet);
	if (!self->IsOpen) return NULL;
// printf("%s(%d): %lld %d\n", FL, Offset, MaxChr);
	Offset += self->Offset;
	if ((Offset < 0) || (Offset >= fio_GetSizeAbs(self))) return NULL;
//	if (Offset >= self->size) return NULL;
	if ((MaxChr + self->BlockSize) > MAX_MEM) return NULL;

	rdseg = False;
	if ((MaxChr + self->BlockSize) > self->BlockMemSize) {
	    int BlockCount = self->BlockCount * 2;
	    if ((BlockCount * self->BlockSize) > MAX_MEM) {
 printf("%s(%d): max. mem %d\n", FL, MAX_MEM);
		return NULL;
	    }

	    self->BlockCount = BlockCount;
	    self->BlockMemSize = self->BlockSize * self->BlockCount;
	    self->BlockBuf = xrealloc(self->BlockBuf, self->BlockMemSize);

	    self->BlockAdr = GetBlockAdr(self, Offset);
	    rdseg = True;
// printf("%s(%d): BlockAdr 0x%04llx\n", FL, self->BlockAdr);

	} else if (Offset < self->BlockAdr) {
	    self->BlockAdr = GetBlockAdr(self, Offset);
	    rdseg = True;
// printf("%s(%d): BlockAdr 0x%04llx\n", FL, self->BlockAdr);

	} else if ((Offset + MaxChr) > (self->BlockAdr + self->BlockMemSize)) {
	    self->BlockAdr = GetBlockAdr(self, Offset);
	    rdseg = True;
// printf("%s(%d): BlockAdr 0x%04llx\n", FL, self->BlockAdr);

	}
	if (rdseg) {
	    if (!TFIOfd_seek(self, self->BlockAdr)) return NULL;
	    int lg = self->BlockMemSize;
	    if (lg > (self->size - self->BlockAdr)) lg = self->size - self->BlockAdr;
	    if (!TFIOfd_read(self, self->BlockBuf, lg)) return NULL;

	    if (self->wrseg) {	//sync BlockBuf  with wrseg
		int64_t BlockAdr = self->BlockAdr;
		char *BlockBuf = self->BlockBuf;
		for (int i=self->BlockCount;i>0;i--) {

		    TWRSeg *wrseg = self->wrseg;
		    while (wrseg && (wrseg->BlockAdr != BlockAdr)) wrseg = wrseg->next;
		    if (wrseg) {
			pnmove(BlockBuf, wrseg->data, self->BlockSize);
		    }
		    BlockBuf += self->BlockSize;
		    BlockAdr += self->BlockSize;
		}
	    }
	}

	return self->BlockBuf + (Offset - self->BlockAdr);
}

// --------------------- write -------------------------------------
static int WRSeg_s(TFIOfd *self, TWRSeg *wrseg, int64_t Offset, const char *s, int size) {
  int off, result;

	off = Offset - wrseg->BlockAdr;
	result = wrseg->BlockSize - off;
	if (size < result) result = size;
	pnmove(wrseg->data + off, s, result);

	if ((self->BlockAdr <= Offset) &&
	    ((self->BlockAdr + self->BlockMemSize) > Offset)) {
	    off = Offset - self->BlockAdr;
// printf("%s(%d): 0x%08x\n", FL, off);
	    pnmove(self->BlockBuf + off, s, result);
	}


	return result;
}

static TWRSeg *GetWRSeg(TFIOfd *self, int64_t Offset) {
  TWRSeg *wrseg;
  int64_t BlockAdr;

	BlockAdr = GetBlockAdr(self, Offset);
	wrseg = self->wrseg;
	while (wrseg && (wrseg->BlockAdr != BlockAdr)) wrseg = wrseg->next;
	if (wrseg) {
	    if (self->BlockSize != wrseg->BlockSize) {
 printf("%s(%d): TODO: wrseg->BlockSize != wrseg->BlockSize\n", FL);
		return NULL;
	    }
	} else {
	    wrseg = AddWRSeg(self, Offset);
	}
	return wrseg;
}

static bool TFIOfd_OverWrite(TFIOfd *self, int64_t Offset, const char *s, int size) {
  TWRSeg *wrseg;
  int64_t max;

	if (!self->IsOpen || !(self->IOfkt & IOfkt_CanWrite)) return False;
	Offset += self->Offset;
	max = fio_GetSizeAbs(self);
	if ((Offset < 0) || (Offset >= max)) return False;

	while (size > 0) {
	    wrseg = GetWRSeg(self, Offset);
// printf("%s(%d): 0x%08x\n", FL, (int)wrseg);
	    if (!wrseg) return False;
	    int n = WRSeg_s(self, wrseg, Offset, s, size);
	    Offset += n;
	    s += n;
	    size -= n;
	}
	return True;
}

static bool TFIOfd_Save(TFIOfd *self, const char *FileName) {
  TWRSeg *wrseg;

	if (!self->IsOpen || !(self->IOfkt & IOfkt_CanWrite) || FileName) return False;
	while (self->wrseg) {
	    wrseg = self->wrseg;
// printf("%s(%d): 0x%08x 0x%08x 0x%08x\n", FL, (int)wrseg, (int)wrseg->BlockAdr, self->BlockSize);
	    if (TFIOfd_seek(self, wrseg->BlockAdr)) {
// printf("%s(%d): %d 0x%02x\n", FL, self->fd, *wrseg->data);
		file_write_ae(self->fd, wrseg->data, wrseg->BlockSize, NULL,
			      self->FileName, EMSG_ERR1);
	    }
	    self->wrseg = wrseg->next;
	    xfree(wrseg);
	}
	if (self->OnWriteEnd) self->OnWriteEnd(self);
	return True;
}

static bool TFIOfd_Close(TFIOfd *self) {

	net_err_clr(&self->ErrRet);
	if (self->IsOpen) {
	    file_close(self->fd, NULL);
	    self->IsOpen = False;
	}
	while (self->wrseg) {
	    TWRSeg *wrseg = self->wrseg;
	    self->wrseg = wrseg->next;
	    xfree(wrseg);
	}
	gp_free(self->gp);
	ppfree(&self->st);

	self->BlockAdr = 0;
	self->size = 0;
	self->Offset = 0;
	self->IOfkt = 0;
	return True;
}

#ifndef _OS_WIN32
static int64_t meminfo_LowTotal() {
  int fd, lg;
  int64_t result;
  char *pc;
  char buf[4096];

	result = 0;
	fd = open("/proc/meminfo", O_RDONLY);
	if (fd > 0) {
	    lg = read(fd, buf, sizeof(buf)-1);
	    close(fd);
	    buf[lg] = 0;
	    pc = buf;
	    while (*pc && (lg > 0)) {
		if (pcncmp("LowTotal:", pc, 9)) {
		    pc += 9;
		    result = pctoint(pc);
		    result = (result << 10); // + 0x00100000;
		    break;
		}
		while (*pc && (*pc != '\n')) {pc++; lg--;}
		if (*pc == '\n') {pc++; lg--;}
	    }
	}
	if (!result) result = 0x01000000;
	return result;
}
#endif /* _OS_WIN32 */


static bool TFIOfd_Open(TFIOfd *self) {
  bool ok;

	TFIOfd_Close(self);
	if (!self->FileName) goto end_msg;

	self->gp = pctogp(self->FileName);
	if (!gp_test_local_drive(self->gp)) goto end_msg;

	self->st = file_stat_gp(self->gp, fsFollowLink | fsGetFileType | fsNoFileName | fsNoLinkName);
	if (!self->st || (self->st == PTR_FAILED)) {
	    if (!self->st) {
		char *fn = get_DisplayFileStr(NULL, self->gp);
		TTC_netError((TTC0*)self, NULL, EMSG_ERR1, "file not found \n%s", fn);
		xfree(fn);
	    }
	    self->st = NULL;
	    goto end_msg;

#ifdef _OS_WIN32
	} else if (!(self->st->st & st_Reg)) {
#else /* _OS_WIN32 */
	} else if (!(self->st->st & st_Reg) && 
		    ((self->st->mode & S_IFMT) != S_IFBLK) &&
		    ((self->st->mode & S_IFMT) != S_IFCHR)) { 
#endif /* !_OS_WIN32 */
	    char *fn = get_DisplayFileStr(NULL, self->gp);
	    TTC_netError((TTC0*)self, NULL, EMSG_ERR1, "file is not a Regular file !!! \n%s", fn);
	    xfree(fn);
	    goto end_msg;
	}

//	if (!(self->st->st & st_ReadOnly)) {
	if (file_chk_access_st(self->st, W_OK)) {
	    self->IOfkt = DEFAULT_IOfkt;
	    ok = file_open_rdwr(self->FileName, &self->ErrRet, &self->fd);
	} else {
	    ok = file_open_rd(self->FileName, &self->ErrRet, &self->fd);
	}

	if (!ok) {
	    net_err_app_4(&self->ErrRet, self->FileName, "open", EMSG_ERR1);
	    goto end_msg;
	}

	if (self->st->st & st_Reg) {
	    self->size = self->st->size;

#ifndef _OS_WIN32
	} else if ((self->st->mode & S_IFMT) == S_IFBLK) {
	    if (ioctl(self->fd, BLKGETSIZE64, &self->size) < 0) {
// printf("%s(%d): 0x%08llx\n", FL, self->size);
		if (ioctl(self->fd, BLKGETSIZE, &self->size) < 0) {
// printf("%s(%d): 0x%08llx\n", FL, self->size);
		    self->size = 0;
		} else self->size <<= 9;
	    }
// printf("%s(%d): 0x%08llx\n", FL, self->size);
	} else if ((self->st->mode & S_IFMT) == S_IFCHR) {
	    if (pccmp(self->FileName, "/dev/mem")) {
		self->size = meminfo_LowTotal();
	    }
#endif /* !_OS_WIN32 */
	}


	self->IsOpen = True;

	int lg = self->BlockMemSize;
	if (lg > self->size) {
	    lg = self->size;
	}
	if (lg > 0) {
	    if (TFIOfd_read(self, self->BlockBuf, lg)) {
		if (self->OnData) self->OnData(self, False);
	    } else {
		TFIOfd_Close(self);
	    }
	}
end_msg:
	if (self->OnReadEnd) self->OnReadEnd(self);
	return self->IsOpen;
}


/* static bool TFIOfd_ReOpen(TFIOfd *self, const char *FileName) {
	ppcchange_cmp(&self->FileName, FileName);
	return TFIOfd_Open(self);
} */
/*
static bool TFIOfd_SaveAs(TFIOfd *self, const char *FileName) {
//	if (!self->IsOpen) return False;

	return False;
}
*/

static bool GetChanged(TFIOfd *self) {
	return self->wrseg != NULL;
}

/*____________________________ Destroy __________________________________*/
static bool TFIOfd__Destroy(TFIOfd *self) {
	TFIOfd_Close(self);
//	gp_free(self->gp);
	xfree(self->BlockBuf);
	return TFIO__Destroy((TFIO*)self);
}

/*____________________________ Create ___________________________________*/
TFIO *TFIO_Create_FD(const char *FileName) {
  TFIOfd *self;

	self = (TFIOfd*)TFIO_Create(FileName, sizeof(TFIOfd), FIOartFD);
	self->destroy = (TDESTROY)TFIOfd__Destroy;

	self->BlockShift = 12;
	self->BlockSize = 1 << self->BlockShift;
	self->BlockCount = 4;
	self->BlockMemSize = self->BlockSize * self->BlockCount;
	self->BlockBuf = xmalloc(self->BlockMemSize);
	self->Open = TFIOfd_Open;
//	self->ReOpen = TFIOfd_ReOpen;
	self->Save = TFIOfd_Save;
//	self->SaveAs = TFIOfd_SaveAs;
	self->GetDataP = TFIOfd_GetDataP;
	self->GetSize = TFIOfd_GetSize;
	self->GetSizeAbs = fio_GetSizeAbs;
	self->SetOffset = fio_SetOffset;
	self->GetOffset = fio_GetOffset;
	self->GetChanged = GetChanged;
	self->OverWrite = TFIOfd_OverWrite;
//	self->Insert = TFIOlp_Insert;
//	self->Delete = TFIOlp_Delete;

	self->IOfkt = DEFAULT_IOfkt;

	return (TFIO*)self;
}

