/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "sysbase.h"
#include "Microstrip.h"
#include <math.h>


static float Q2(float x) {return x*x;}
static float Q3(float x) {return x*x*x;}

static float Eeff_Hammerstad_Jensen(float W, float H, float Er) {
  float u, u4, b, a, Eeff;

	u = W/H;
	u4 = u*u*u*u;
	b = 0.564*pow((Er-0.9)/(Er+3), 0.053);
	a = 1 + 0.02040816*log( (u4+Q2(u/52)) / (u4+0.432)) +
		0.05347594*log(1+Q3(u/18.1));
	Eeff = (Er+1)/2 + (Er-1)/2 * pow(1+10*H/W, -a*b);
	return Eeff;
}

static float StripThickKorr(float W, float H, float T, float Er) {
  float w1, wr;
	if (T <= 0) return W;
	w1 = T/M_PI * log(4*M_E/sqrt(Q2(T/H)+(M_1_PI/(W/T+1.1))));
	wr = 0.5*w1*(1+1/Er);
	return wr;
}

static float Z_Hammerstad_Jensen(float W, float H, float Er) {
  float fu, Z;
	fu = 6 + (2*M_PI-6)*exp(-pow(30.666*H/W, 0.7528));
	Z = 60/sqrt(Er) * log(fu*H/W + sqrt(1+Q2(2*H/W)));
	return Z;
}

static float mat_K(float k1, float k2) {
  float res;
	if (k1 < 0.70710678) {
	    res = M_PI / log(2*(1+sqrt(k2))/(1-sqrt(k2)));
	} else {
	    res = log(2*(1+sqrt(k1))/(1-sqrt(k1))) / M_PI;
	}
	return res;
}

static void Math_CPW(TMicrostripDat *md, float *pEeff, float *pZ) {
  float Z, f, k1, k2, ke, Eeff, Eefft;
  float W = md->W;
  float H = md->H;
  float T = md->T;
  float Er = md->Er;
  float s = md->G;
  float s2 = 2*s;

	k1 = W / (W+s2);
	k2 = sinh(M_PI*W/4/H) / sinh(M_PI*(W+s2)/4/H);
	Eeff = 1 + (Er-1)/2 * mat_K(k2, sqrt(1-k2*k2)) * mat_K(sqrt(1-k1*k1), k1);
	f = 1.25*T/M_PI*(1+log(4*M_PI*W/T));
	ke = k1+(1-k1*k1)*f/s2;
	Eefft = Eeff-0.7*(Eeff-1)*T/s /
		(mat_K(k1, sqrt(1-k1*k1)) + 0.7*T/s );
	Z = 30*M_PI/sqrt(Eefft) * mat_K(sqrt(1-ke*ke), ke);
	*pZ = Z;
	*pEeff = Eeff;
}



void Microstrip_math(TMicrostripDat *md) {
  float f, Z, pfm, Eeff, t, Z_CPW, Eeff_CPW;
  float W0 = md->W;
  float W = md->W;
  float G = md->G;
  float H = md->H;
  float T = md->T;
  float Er = md->Er;

	Eeff = Eeff_Hammerstad_Jensen(W0, H, Er);
	W += StripThickKorr(W0, H, T, Eeff);
	Z = Z_Hammerstad_Jensen(W, H, Eeff);

	if (md->Coplanar) {
	    Math_CPW(md, &Eeff_CPW, &Z_CPW);
	    f = 1-0.49*exp(-0.96 * G * 2.3 / H);
	    Eeff *= sqrt(f);
	    Z *= f; //*1.06;

	    if (md->GroundPlane) {
		if (Eeff < Eeff_CPW) Eeff = Eeff_CPW;
		if (Z > Z_CPW) Z = Z_CPW;
	    } else {
		Eeff = Eeff_CPW;
		Z = Z_CPW;
	    }
	    t = sqrt(Eeff * sqrt(W0 / W)) / 0.2992;
	}

	md->Eeff = Eeff;
	md->Eeff_W = W;
	md->Z = Z;

	t = sqrt(Eeff * sqrt(W0 / W)) / 0.2992;
	md->nSm = t;

	pfm = t * 1000 / Z;
	md->pFm = pfm;

	f = pfm * Z * Z / 1000000;
	md->uHm = f;

}

void Microstrip_math_Freq(TMicrostripDat *md) {
  float A, st, f, k;
  float Freq = md->Freq;
  float T = md->T;
  float W0 = md->W;

	md->dbm = 0.09055 * Freq * tan(md->Er_tan) * sqrt(md->Eeff);

	A = W0 * T;
	st = 0.132 / sqrt(Freq);
	if ((st < W0) && (st < T)) {
	    A -= (W0-st)*(T-st);
	}
	md->Ohmm = 0.01786 / A;

	md->OhmPm = 9999999;
	if (Freq > 0) {
	    f = 6.28 * Freq * md->pFm * 0.000001;
	    k = 0.000001;
	    if (md->Er > 1) {
		k =  (md->Eeff-1) / (md->Er-1);
	    }
	    md->OhmPm = 1 / (k * f * tan(md->Er_tan));
	}
}

void Microstrip_math_L(TMicrostripDat *md) {
  float L = md->L / 1000;
	md->nS = md->nSm * L;
	md->pF = md->pFm * L;
	md->uH = md->uHm * L;
	md->db = md->dbm * L;
	md->Ohm = md->Ohmm * L;
	md->OhmP = md->OhmPm / L;
}
