/*
 Copyright (C) 2010-2014 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"

static char *pack_scan_path;
static const char *pack_idx_file;
static const char *package_name;
static char *pack_Version;
static char *pack_URL;
static char *pack_rpm;
//static char *path0, *name0;
static TpcIni *ini;

static void data_save(const char *name, const char *rpath, const char *data, const char *D) {
  char buf[1024];
	if (!name || !rpath || !data) return;
	if (!ini) {
	    ini = TpcIni_Create(pack_idx_file);
	    if (!pack_Version) pack_Version = TpcIni_RDstrdup(ini, package_name, "Version", NULL);
	    if (!pack_URL) pack_URL = TpcIni_RDstrdup(ini, package_name, "URL", NULL);
	    if (!pack_rpm) pack_URL = TpcIni_RDstrdup(ini, package_name, "rpm", NULL);
	    TpcIni_DelGruppeDat(ini, package_name);
	    if (pack_Version) TpcIni_WR(ini, package_name, "Version", pack_Version);
	    if (pack_URL) TpcIni_WR(ini, package_name, "URL", pack_URL);
	    if (pack_rpm) TpcIni_WR(ini, package_name, "rpm", pack_rpm);
	}
	snprintf(buf, sizeof(buf), "%s.%s/%s", D, name, rpath);
	TpcIni_WR(ini, package_name, buf, data);
}

static void add_ini_str(char **ppc, const char *pc) {
	ppcadd_n(ppc, 3, "\"", pc, "\"");
}

static void scan_exec(TDirFileList *dfl, const char *rpath, const char *D) {
  char *pc_exec = NULL;
  char *pc_so = NULL;
  char *pc_sh = NULL;
  char *pc_sh_perl = NULL;

	while (dfl) {
	    switch (dfl->st->FileType) {
	    case stft_Exec:
		add_ini_str(&pc_exec, dfl->fn);
		break;
	    case stft_SO:
		if (!(dfl->st->st & st_Link)) {
		    add_ini_str(&pc_so, dfl->fn);
		}
		break;
	    case stft_Shell:
		add_ini_str(&pc_sh, dfl->fn);
		break;
	    case stft_sh_perl:
		add_ini_str(&pc_sh_perl, dfl->fn);
		break;
	    }
	    dfl = dfl->next;
	}
	data_save("exec____", rpath, pc_exec, D);
	data_save("so______", rpath, pc_so, D);
	data_save("sh______", rpath, pc_sh, D);
	data_save("sh_perl_", rpath, pc_sh_perl, D);
}

static void scan_sub_path1(Tstat64 *st, const char *xpart, const char *D) {
  const char *fn;
  char *rpath, *path;
  TDirList *dl;
  TDirFileList *dfl;

	if (st && !(st->st & st_Dir)) return;
	if (st && (st->st & st_Link)) D = "L";
	fn = Get_ExFileDat_pc(st, exfdt_FileName);
	rpath = path_add(xpart, fn);

	path = path_add(pack_scan_path, rpath);
	ppcadd(&path, "/");

	dl = TDirList_Create(path);
	if (TDirList_Open(dl, dlobef_Rescan)) {
	    TDirList_Open_Wait(dl, dlrunstat_ReScanEnd);
	    TDirFileList_Sort(&dl->dfl, 0);
	    scan_exec(dl->dfl, rpath, D);

	    dfl = dl->dfl;
	    while (dfl) {
		if (!pccmp(dfl->fn, "..")) {
		    scan_sub_path1(dfl->st, rpath, D);
		}
		dfl = dfl->next;
	    }
	}
	Tfree(dl);
	xfree(path);

	xfree(rpath);
}

static const char *Help_Msg = {
  "\n"
  "Usage: packages_idx [options] scan_path\n"
  "\n"
  "  -d                 package index file\n"
  "                         (default: /usr/src/packages/pack_idx.txt)\n"
  "  -p                 package name\n"
  "  -r                 package rpm (-r i386/bzip2-1.0.6-0.i386.rpm)\n"
  "  -u                 package URL\n"
  "  -v                 package version\n"
  "  -h, --help         Prints out this screen\n"
  "\n"
  "    packages_idx /var/tmp/bzip2-1.0.6\n"
  "    packages_idx -p bzip2 /var/tmp/bzip2-1.0.6\n\n"
  "  ALDccp: packages_idx version "ALD_VERSION"  (build %d  %s)\n"
};

int main (int argc, char **argv) {
  int i;
  char *val;
  bool err = False;
  bool help = False;

	pack_scan_path = NULL;
	pack_idx_file = "/usr/src/packages/pack_idx.txt";
	package_name = NULL;
	pack_Version = NULL;
	pack_URL = NULL;
	pack_rpm = NULL;
	ini = NULL;

	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'd':
		    i++;
		    if (i<argc) pack_idx_file = argv[i]; else err = True;
		    break;
		case 'p':
		    i++;
		    if (i<argc) package_name = argv[i]; else err = True;
		    break;
		case 'r':
		    i++;
		    if (i<argc) pack_rpm = argv[i]; else err = True;
		    break;
		case 's':
		    i++;
		    if (i<argc) pack_scan_path = argv[i]; else err = True;
		    break;
		case 'u':
		    i++;
		    if (i<argc) pack_URL = argv[i]; else err = True;
		    break;
		case 'v':
		    i++;
		    if (i<argc) pack_Version = argv[i]; else err = True;
		    break;
		case 'h':
		    help = True;
		    break;
		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
		    } else err = True;
		    break;
		default: 
		   err = True;
		}
	    } else {
		if (pack_scan_path) err = True;
		pack_scan_path = val;
	    }
	    i++;
	}

	if (help || err || !pack_scan_path) {
	    printf(Help_Msg, ALD_BUILD_V, __DATE__);
	    exit(err);
	}

	initapp(app_NoIni, argc, argv);

	hGetFileHeadInfo = &hGetFileHeadInfo_buildin;
	ThDirList_Reg(1, &hdirlist_ld_File);

  Tstat64 *st;
	st = file_stat_fn(pack_scan_path, 0);
	if (!st || !(st->st & st_Dir)) {
	    printf("ERROR: path not found \"%s\"\n", pack_scan_path);
	    exit(1);
	}
	if (!package_name) package_name = GetFileName_pc(pack_scan_path);

	xfree(st);

	scan_sub_path1(NULL, NULL, "D");

	TpcIni_SaveAndFree(ini);
}
